extern crate eui48;
extern crate byteorder;
extern crate bitfield;

use byteorder::{ByteOrder, LittleEndian};
use eui48::{MacAddress, Eui48};
use std::net::{IpAddr, Ipv4Addr};
use bitfield::bitfield;


/* ethernet */
const ETH_ALEN: usize = 6;
const ETH_TLEN: usize = 2;
const ETHER_HDRLEN: usize = 14;


#[derive(Debug, Clone)]
pub struct EtherHeader {
//   pub ether_dhost: MacAddress,
//   pub ether_shost: MacAddress,
   pub ether_dhost: String,
   pub ether_shost: String, 
   pub ether_type: i32,
}

pub fn ethernet_handler ( packet_data: &[u8] ) -> EtherHeader {
    let mut _ether_dhost: [u8; ETH_ALEN] = [0; ETH_ALEN];
    let mut _ether_shost: [u8; ETH_ALEN] = [0; ETH_ALEN];
    let mut _ether_type:  u16 = 0;

    _ether_dhost.clone_from_slice(&packet_data[0..ETH_ALEN]);
    
    //println!("{:?}", (&(_ether_dhost).to_owned()));
    _ether_shost.clone_from_slice(&packet_data[ETH_ALEN..ETH_ALEN*2]);
    _ether_type = LittleEndian::read_u16(&packet_data[ETH_ALEN*2..(ETH_ALEN*2)+ETH_TLEN]);
    
    EtherHeader {
        ether_dhost: (MacAddress::new(_ether_dhost as Eui48).to_hex_string()),
        ether_shost: (MacAddress::new(_ether_shost as Eui48).to_hex_string()),
        ether_type:  _ether_type as i32,
    }
}

/* ip */
#[derive(Debug,Copy, Clone)]
pub struct IpV4Header {
   pub ip_version:                     u32,
   pub ip_ihl:                         u32,
   pub ip_dscp:                        u32,
   pub ip_ecn:                         u32,
   pub ip_total_length:                u32,
   pub ip_identification:              u32,
   pub ip_df:                          u32,
   pub ip_mf:                          u32,
   pub ip_fragment_offset:             u32,
   pub ip_time_to_live:                u32,
   pub ip_protocol:                    u32,
   pub ip_header_checksum:             u32,
   pub ip_source_address:              IpAddr,
   pub ip_destination_address:         IpAddr,

}

bitfield! {
    struct BitfieldIpV4Header(MSB0 [u8]);
    impl Debug;
    u32;
    get_version, _: 3, 0;
    get_ihl, _: 7, 4;
    get_dscp, _: 13, 8;
    get_ecn, _: 15, 14;
    get_total_length, _: 31, 16;
    get_identification, _: 47, 32;
    get_df, _: 49;
    get_mf, _: 50;
    get_fragment_offset, _: 63, 51;
    get_time_to_live, _: 71, 64;
    get_protocol, _: 79, 72;
    get_header_checksum, _: 95, 80;
    u8, get_source_address, _: 103, 96, 4;
    u32, into Ipv4Addr, get_destination_address, _: 159, 128;
}

impl<T: AsRef<[u8]> + AsMut<[u8]>> BitfieldIpV4Header<T> {
    fn get_source_as_ip_addr(&self) -> Ipv4Addr {
        let mut src = [0; 4];
        for (i, src) in src.iter_mut().enumerate() {
            *src = self.get_source_address(i);
        }
        src.into()
    }
}

pub fn ip_handler ( packet_data: &[u8] ) -> IpV4Header {
    let (_head, tail) = packet_data.split_at(ETHER_HDRLEN);
    let (raw_hdr, _) = tail.split_at(20);
    let mut _tail: [u8; 20] = [0; 20];
    _tail.copy_from_slice(raw_hdr);
    
    let ip_header = BitfieldIpV4Header(_tail);
        
    IpV4Header {
    ip_version:             ip_header.get_version(),                     
    ip_ihl:                 ip_header.get_ihl(),                         
    ip_dscp:                ip_header.get_dscp(),                        
    ip_ecn:                 ip_header.get_ecn(),                         
    ip_total_length:        ip_header.get_total_length(),                
    ip_identification:      ip_header.get_identification(),              
    ip_df:                  ip_header.get_df().into(),                   
    ip_mf:                  ip_header.get_mf().into(),                   
    ip_fragment_offset:     ip_header.get_fragment_offset(),             
    ip_time_to_live:        ip_header.get_time_to_live(),                
    ip_protocol:            ip_header.get_protocol(),                    
    ip_header_checksum:     ip_header.get_header_checksum(),             
    ip_source_address:      IpAddr::V4(ip_header.get_source_as_ip_addr()),    
    ip_destination_address: IpAddr::V4(ip_header.get_destination_address()), 
    }
}

/* tcp */
#[derive(Debug,Copy,Clone)]
pub struct TcpHeader {
   pub source_port:              u32,
   pub destination_port:         u32,
   pub seq_num:                  u32,
   pub ack_num:                  u32,
   pub data_offset:              u32,
   pub reserved:                 u32,
   pub ns:                       u32,
   pub cwr:                      u32,
   pub ece:                      u32,
   pub urg:                      u32,
   pub ack:                      u32,
   pub psh:                      u32,
   pub rst:                      u32,
   pub syn:                      u32,
   pub fin:                      u32,
   pub window_size:              u32,
   pub checksum:                 u32,
   pub urgent_pointer:           u32,
}


bitfield! {
    struct BitfieldTcpHeader (MSB0 [u8]);
    u32;
    get_source_port, _:             15, 0;
    get_destination_port, _:        31,16;
    get_seq_num, _:                 63,32;
    get_ack_num, _:                 95,64;
    get_data_offset, _:             99,96;
    get_reserved, _:                102,100;
    get_ns, _:                      103;
    get_cwr, _:                     104;
    get_ece, _:                     105;
    get_urg, _:                     106;
    get_ack, _:                     108;
    get_psh, _:                     108;
    get_rst, _:                     109;
    get_syn, _:                     110;
    get_fin, _:                     111;
    get_window_size, _:             127,112;
    get_checksum, _:                143,128;
    get_urgent_pointer, _:             159,144;
}

pub fn tcp_handler ( ip_hlen: u32, packet_data: &[u8] ) ->TcpHeader {
    let (_head, tail) = packet_data.split_at(ETHER_HDRLEN+ip_hlen as usize * 4);
    let (raw_hdr, _) = tail.split_at(20);
    let mut _tail: [u8; 20] = [0; 20];
    _tail.copy_from_slice(raw_hdr);
    let tcp_header = BitfieldTcpHeader(_tail);
    
    TcpHeader {
    source_port:           tcp_header.get_source_port(),   
    destination_port:      tcp_header.get_destination_port(),        
    seq_num:               tcp_header.get_seq_num(),                 
    ack_num:               tcp_header.get_ack_num(),                 
    data_offset:           tcp_header.get_data_offset(),             
    reserved:              tcp_header.get_reserved(),                
    ns:                    tcp_header.get_ns().into(),                      
    cwr:                   tcp_header.get_cwr().into(),                     
    ece:                   tcp_header.get_ece().into(),                     
    urg:                   tcp_header.get_urg().into(),                     
    ack:                   tcp_header.get_ack().into(),                     
    psh:                   tcp_header.get_psh().into(),                     
    rst:                   tcp_header.get_rst().into(),                     
    syn:                   tcp_header.get_syn().into(),                     
    fin:                   tcp_header.get_fin().into(),                     
    window_size:           tcp_header.get_window_size(),             
    checksum:              tcp_header.get_checksum(),                
    urgent_pointer:        tcp_header.get_urgent_pointer(),          
    }    


}

/* payload */
pub fn payload_handler ( ip_hlen: u32, data_offset: u32, packet_data : &[u8] ) -> Option<Vec<u8>> {
    let (_head, tail)= packet_data.split_at(ETHER_HDRLEN+ip_hlen as usize * 4+data_offset as usize * 4);
    Some(tail.to_vec())    
}


